SpriteAnimator = class("SpriteAnimator")

function SpriteAnimator:created(fps, shouldNoSmooth)
    self.defaultFPS = fps
    self.fps = fps
    
    self.animations = {}
    self.frameNumber = 1
    
    self.timeOffset = time.elapsed
    self.repeatNum = nil
    
    self.currentAnimation = nil
    self.currentAnimationName = nil
    self.playingAnimation = false
    self.hasEnded = false
    self.lock = false
    self.gotoframe = nil
    self.smooth = true
    if shouldNoSmooth then
        self.smooth = false
    end
    
    self.canEnd = false
    self.animationFinish = false
    self.editing = false
    
    -- This is for animation sheets
    self.animationSheetImg = nil
    self.animationSheet = nil
    self.isVertical = nil
    self.sheetSize = nil
    
    
    self.onCompleteFunction = nil
end

function SpriteAnimator:setSheet(sheetImg, size, isVertical)
    
    -- puts the sheet into an atlas and divide the atlas
    local aniSheet = image.read(sheetImg)
    self.animationSheetImg = aniSheet
    self.animationSheet = aniSheet.atlas
    self.animationSheet:setWithCellSize(size.x, size.y)
    self.sheetSize = vec2(aniSheet.width / size.x, aniSheet.height/size.y)
    
    -- this is if your animation sheet is downwards and not from left to right
    self.isVertical = isVertical
    if self.isVertical == nil then
        self.isVertical = false
    end
    return self
end

function SpriteAnimator:add(name, ...)
    self.animations[name] = SpriteAnimation(...)
    return self
end

function SpriteAnimator:addFrom(name, from, to)
    
    local animation = SpriteAnimation()
    self.animations[name] = animation
    animation.sheetImg = self.animationSheetImg
    
    local curPos = from
    repeat
        local frameImg = self.animationSheet["c"..math.tointeger(curPos.y).."r"..math.tointeger(curPos.x)]
        table.insert(animation.frames, frameImg)
        
        if not self.isVertical then
            curPos.y = curPos.y + 1
            if curPos.y > self.sheetSize.x then
                curPos.y = 1
                curPos.x = curPos.x + 1
            end
        else
            curPos.x = curPos.x + 1
            if curPos.x > self.sheetSize.y then
                curPos.x = 1
                curPos.y = curPos.y + 1
            end
        end
            
    until (curPos == to)
    local frameImg = self.animationSheet["c"..math.tointeger(curPos.y).."r"..math.tointeger(curPos.x)]
    table.insert(animation.frames, frameImg)
    
    
    return self
end

function SpriteAnimator:pause()
    self.playingAnimation = false
    return self
end

function SpriteAnimator:resume() 
    if not self.playingAnimation then
        self:goToFrame(self.frameNumber + 1, true)
        self:checkGoto()
        self.frameNumber = (math.floor((time.elapsed - self.timeOffset) * self.fps) % #self.currentAnimation.frames) + 1
        self.playingAnimation = true
    end
    return self
end

-- Allow the animator to go to a specific frame
function SpriteAnimator:goToFrame(frameNum, wrap)
    self.gotoframe = ((frameNum - 1) % #self.currentAnimation.frames) + 1
    if wrap == nil and frameNum > #self.currentAnimation.frames then
        warning("You went over the number of frames without wrapping: it did it anyways")
    end
end 

function SpriteAnimator:getAnimation(name)
    return self.animations[name]
end

function SpriteAnimator:getCurrentAnimation()
    return self.currentAnimation
end

function SpriteAnimator:play(name, keepFrameNumber)
    -- Check if able to change the animation then reset all the property of the animator
    if not self.lock and (self.currentAnimationName ~= name or self.animationFinish) then
        self.playingAnimation = true
        self.currentAnimation = self.animations[name]
        self.currentAnimationName = name
        self.entity.sprite = self.currentAnimation.frames[1]
        self.isStart = true
        if not keepFrameNumber then
            self.gotoframe = 1
        else
            self.gotoframe = self.frameNumber 
        end
        self.hasEnded = false
        self.animationFinish = false
        self.fps = self.defaultFPS
        self.repeatNum = nil
        self.editing = true
        self.onCompleteFunction = nil
        self.calledCompleteFunction = nil
    end
    return self
end

function SpriteAnimator:onComplete(func)
    self.onCompleteFunction = func
end

function SpriteAnimator:loop(num)
    if self.editing then
        self.repeatNum = num
    end
    return self
end

function SpriteAnimator:setFPS(fps)
    if self.editing then
        self.fps = fps
    end
    return self
end

function SpriteAnimator:setSmooth(shouldSmooth)
    self.smooth = shouldSmooth
end

function SpriteAnimator:update()
    self.editing = false
    
    --When the looping of the animation stops
    if self.repeatNum == 0 then
        self.animationFinish = true
    end
    
    -- call the onComplete function
    if self.animationFinish and self.onCompleteFunction ~= nil and not self.calledCompleteFunction then
        self.onCompleteFunction()
        self.calledCompleteFunction = true
    end
    
    -- loops the animation if it is playing
    if self.playingAnimation  then
        
        self.currentAnimation.sheetImg.smooth = self.smooth
        self:checkGoto()
        
        -- Change the animation to the right frame 
        self.frameNumber = (math.floor((time.elapsed - self.timeOffset) * self.fps) % #self.currentAnimation.frames) + 1
        
        -- check if this is the last frame of the animation
        if self.hasEnded then
            self.hasEnded = false
            self.canEnd = false
        elseif self.frameNumber == 1 and self.canEnd and not self.isStart then
            self.hasEnded = true
        end
        
        if self.frameNumber == #self.currentAnimation.frames then
            self.canEnd = true
            self.isStart = false
        end
        
        -- if the last frame then decrease the loop value
        if self.hasEnded and self.repeatNum then
            self.repeatNum = self.repeatNum - 1
        end
        
        -- if the loop has ended then stop animation
        if self.repeatNum == 0 then
            self.frameNumber = #self.currentAnimation.frames
            self.playingAnimation = false
        end
        
        -- set the sprite to the current frame
        self.entity.sprite = self.currentAnimation.frames[self.frameNumber]
    end
end

function SpriteAnimator:checkGoto()
    if self.gotoframe then
        self.timeOffset = time.elapsed - ((self.gotoframe - 1.0) / self.fps)
        self.gotoframe = nil
    end
end
